@@ -1,6 +1,10 @@
 # -*- coding: utf-8 -*-
 """Display formatters.
 
+Inheritance diagram:
+
+.. inheritance-diagram:: IPython.core.formatters
+   :parts: 3
 
 Authors:
 
@@ -8,7 +12,7 @@ Authors:
 * Brian Granger
 """
 #-----------------------------------------------------------------------------
-# Copyright (c) 2010, IPython Development Team.
+# Copyright (C) 2010-2011, IPython Development Team.
 #
 # Distributed under the terms of the Modified BSD License.
 #
@@ -22,13 +26,17 @@ Authors:
 # Stdlib imports
 import abc
 import sys
+import warnings
 # We must use StringIO, as cStringIO doesn't handle unicode properly.
 from StringIO import StringIO
 
 # Our own imports
 from IPython.config.configurable import Configurable
 from IPython.lib import pretty
-from IPython.utils.traitlets import Bool, Dict, Int, Unicode, CUnicode, ObjectName
+from IPython.utils.traitlets import (
+    Bool, Dict, Integer, Unicode, CUnicode, ObjectName, List,
+)
+from IPython.utils.py3compat import unicode_to_str
 
 
 #-----------------------------------------------------------------------------
@@ -40,10 +48,36 @@ class DisplayFormatter(Configurable):
 
     # When set to true only the default plain text formatter will be used.
     plain_text_only = Bool(False, config=True)
-
+    def _plain_text_only_changed(self, name, old, new):
+        warnings.warn("""DisplayFormatter.plain_text_only is deprecated.
+        
+        Use DisplayFormatter.active_types = ['text/plain']
+        for the same effect.
+        """, DeprecationWarning)
+        if new:
+            self.active_types = ['text/plain']
+        else:
+            self.active_types = self.format_types
+    
+    active_types = List(Unicode, config=True,
+        help="""List of currently active mime-types to display.
+        You can use this to set a white-list for formats to display.
+        
+        Most users will not need to change this value.
+        """)
+    def _active_types_default(self):
+        return self.format_types
+    
+    def _active_types_changed(self, name, old, new):
+        for key, formatter in self.formatters.items():
+            if key in new:
+                formatter.enabled = True
+            else:
+                formatter.enabled = False
+    
     # A dict of formatter whose keys are format types (MIME types) and whose
     # values are subclasses of BaseFormatter.
-    formatters = Dict(config=True)
+    formatters = Dict()
     def _formatters_default(self):
         """Activate the default formatters."""
         formatter_classes = [
@@ -51,13 +85,14 @@ class DisplayFormatter(Configurable):
             HTMLFormatter,
             SVGFormatter,
             PNGFormatter,
+            JPEGFormatter,
             LatexFormatter,
             JSONFormatter,
             JavascriptFormatter
         ]
         d = {}
         for cls in formatter_classes:
-            f = cls(config=self.config)
+            f = cls(parent=self)
             d[f.format_type] = f
         return d
 
@@ -72,8 +107,10 @@ class DisplayFormatter(Configurable):
         * text/html
         * text/latex
         * application/json
+        * application/javascript
         * image/png
-        * immage/svg+xml
+        * image/jpeg
+        * image/svg+xml
 
         Parameters
         ----------
@@ -84,48 +121,49 @@ class DisplayFormatter(Configurable):
             format data dict. If this is set *only* the format types included
             in this list will be computed.
         exclude : list or tuple, optional
-            A list of format type string (MIME types) to exclue in the format
+            A list of format type string (MIME types) to exclude in the format
             data dict. If this is set all format types will be computed,
             except for those included in this argument.
 
         Returns
         -------
-        format_dict : dict
-            A dictionary of key/value pairs, one or each format that was
+        (format_dict, metadata_dict) : tuple of two dicts
+        
+            format_dict is a dictionary of key/value pairs, one of each format that was
             generated for the object. The keys are the format types, which
             will usually be MIME type strings and the values and JSON'able
             data structure containing the raw data for the representation in
             that format.
+            
+            metadata_dict is a dictionary of metadata about each mime-type output.
+            Its keys will be a strict subset of the keys in format_dict.
         """
         format_dict = {}
-
-        # If plain text only is active
-        if self.plain_text_only:
-            formatter = self.formatters['text/plain']
-            try:
-                data = formatter(obj)
-            except:
-                # FIXME: log the exception
-                raise
-            if data is not None:
-                format_dict['text/plain'] = data
-            return format_dict
+        md_dict = {}
 
         for format_type, formatter in self.formatters.items():
-            if include is not None:
-                if format_type not in include:
-                    continue
-            if exclude is not None:
-                if format_type in exclude:
-                    continue
+            if include and format_type not in include:
+                continue
+            if exclude and format_type in exclude:
+                continue
+            
+            md = None
             try:
                 data = formatter(obj)
             except:
                 # FIXME: log the exception
                 raise
+            
+            # formatters can return raw data or (data, metadata)
+            if isinstance(data, tuple) and len(data) == 2:
+                data, md = data
+            
             if data is not None:
                 format_dict[format_type] = data
-        return format_dict
+            if md is not None:
+                md_dict[format_type] = md
+            
+        return format_dict, md_dict
 
     @property
     def format_types(self):
@@ -315,9 +353,9 @@ class BaseFormatter(Configurable):
 class PlainTextFormatter(BaseFormatter):
     """The default pretty-printer.
 
-    This uses :mod:`IPython.external.pretty` to compute the format data of
+    This uses :mod:`IPython.lib.pretty` to compute the format data of
     the object. If the object cannot be pretty printed, :func:`repr` is used.
-    See the documentation of :mod:`IPython.external.pretty` for details on
+    See the documentation of :mod:`IPython.lib.pretty` for details on
     how to write pretty printers.  Here is a simple example::
 
         def dtype_pprinter(obj, p, cycle):
@@ -353,30 +391,30 @@ class PlainTextFormatter(BaseFormatter):
     verbose = Bool(False, config=True)
 
     # The maximum width.
-    max_width = Int(79, config=True)
+    max_width = Integer(79, config=True)
 
     # The newline character.
     newline = Unicode('\n', config=True)
-    
+
     # format-string for pprinting floats
     float_format = Unicode('%r')
     # setter for float precision, either int or direct format-string
     float_precision = CUnicode('', config=True)
-    
+
     def _float_precision_changed(self, name, old, new):
         """float_precision changed, set float_format accordingly.
-        
+
         float_precision can be set by int or str.
         This will set float_format, after interpreting input.
         If numpy has been imported, numpy print precision will also be set.
-        
+
         integer `n` sets format to '%.nf', otherwise, format set directly.
-        
+
         An empty string returns to defaults (repr for float, 8 for numpy).
-        
+
         This parameter can be set via the '%precision' magic.
         """
-        
+
         if '%' in new:
             # got explicit format string
             fmt = new
@@ -393,7 +431,7 @@ class PlainTextFormatter(BaseFormatter):
                 raise ValueError("Precision must be int or format string, not %r"%new)
             except AssertionError:
                 raise ValueError("int precision must be non-negative, not %r"%i)
-            
+
             fmt = '%%.%if'%i
             if 'numpy' in sys.modules:
                 # set numpy precision if it has been imported
@@ -408,7 +446,7 @@ class PlainTextFormatter(BaseFormatter):
                 numpy.set_printoptions(precision=8)
         self.float_format = fmt
 
-    # Use the default pretty printers from IPython.external.pretty.
+    # Use the default pretty printers from IPython.lib.pretty.
     def _singleton_printers_default(self):
         return pretty._singleton_pprinters.copy()
 
@@ -436,7 +474,7 @@ class PlainTextFormatter(BaseFormatter):
             # ensure that stream does not get a mix of unicode and bytestrings,
             # or it will cause trouble.
             printer = pretty.RepresentationPrinter(stream, self.verbose,
-                self.max_width, self.newline.encode(),
+                self.max_width, unicode_to_str(self.newline),
                 singleton_pprinters=self.singleton_printers,
                 type_pprinters=self.type_printers,
                 deferred_pprinters=self.deferred_printers)
@@ -454,7 +492,7 @@ class HTMLFormatter(BaseFormatter):
     this.
 
     The return value of this formatter should be a valid HTML snippet that
-    could be injected into an existing DOM. It should *not* include the 
+    could be injected into an existing DOM. It should *not* include the
     ```<html>`` or ```<body>`` tags.
     """
     format_type = Unicode('text/html')
@@ -495,6 +533,22 @@ class PNGFormatter(BaseFormatter):
     print_method = ObjectName('_repr_png_')
 
 
+class JPEGFormatter(BaseFormatter):
+    """A JPEG formatter.
+
+    To define the callables that compute the JPEG representation of your
+    objects, define a :meth:`_repr_jpeg_` method or use the :meth:`for_type`
+    or :meth:`for_type_by_name` methods to register functions that handle
+    this.
+
+    The return value of this formatter should be raw JPEG data, *not*
+    base64 encoded.
+    """
+    format_type = Unicode('image/jpeg')
+
+    print_method = ObjectName('_repr_jpeg_')
+
+
 class LatexFormatter(BaseFormatter):
     """A LaTeX formatter.
 
@@ -504,7 +558,8 @@ class LatexFormatter(BaseFormatter):
     this.
 
     The return value of this formatter should be a valid LaTeX equation,
-    enclosed in either ```$``` or ```$$```.
+    enclosed in either ```$```, ```$$``` or another LaTeX equation
+    environment.
     """
     format_type = Unicode('text/latex')
 
@@ -530,7 +585,7 @@ class JavascriptFormatter(BaseFormatter):
     """A Javascript formatter.
 
     To define the callables that compute the Javascript representation of
-    your objects, define a :meth:`_repr_javascript_` method or use the 
+    your objects, define a :meth:`_repr_javascript_` method or use the
     :meth:`for_type` or :meth:`for_type_by_name` methods to register functions
     that handle this.
 
@@ -546,6 +601,7 @@ FormatterABC.register(PlainTextFormatter)
 FormatterABC.register(HTMLFormatter)
 FormatterABC.register(SVGFormatter)
 FormatterABC.register(PNGFormatter)
+FormatterABC.register(JPEGFormatter)
 FormatterABC.register(LatexFormatter)
 FormatterABC.register(JSONFormatter)
 FormatterABC.register(JavascriptFormatter)
@@ -562,8 +618,10 @@ def format_display_data(obj, include=None, exclude=None):
     * text/html
     * text/latex
     * application/json
+    * application/javascript
     * image/png
-    * immage/svg+xml
+    * image/jpeg
+    * image/svg+xml
 
     Parameters
     ----------
@@ -594,3 +652,4 @@ def format_display_data(obj, include=None, exclude=None):
         include,
         exclude
     )
+
