@@ -19,12 +19,12 @@ Authors:
 # Imports
 #----------------------------------------------------------------------
 
-from __future__ import print_function
-
 import logging
 import sys
+import time
 
-from datetime import datetime, timedelta
+from collections import deque
+from datetime import datetime
 from random import randint, random
 from types import FunctionType
 
@@ -40,11 +40,12 @@ from zmq.eventloop import ioloop, zmqstream
 from IPython.external.decorator import decorator
 from IPython.config.application import Application
 from IPython.config.loader import Config
-from IPython.utils.traitlets import Instance, Dict, List, Set, Int, Enum, CBytes
+from IPython.utils.traitlets import Instance, Dict, List, Set, Integer, Enum, CBytes
+from IPython.utils.py3compat import cast_bytes
 
-from IPython.parallel import error
+from IPython.parallel import error, util
 from IPython.parallel.factory import SessionFactory
-from IPython.parallel.util import connect_logger, local_logger, asbytes
+from IPython.parallel.util import connect_logger, local_logger
 
 from .dependency import Dependency
 
@@ -66,18 +67,18 @@ def plainrandom(loads):
 
 def lru(loads):
     """Always pick the front of the line.
-    
+
     The content of `loads` is ignored.
-    
+
     Assumes LRU ordering of loads, with oldest first.
     """
     return 0
 
 def twobin(loads):
     """Pick two at random, use the LRU of the two.
-    
+
     The content of loads is ignored.
-    
+
     Assumes LRU ordering of loads, with oldest first.
     """
     n = len(loads)
@@ -87,7 +88,7 @@ def twobin(loads):
 
 def weighted(loads):
     """Pick two at random using inverse load as weight.
-    
+
     Return the less loaded of the two.
     """
     # weight 0 a million times more than 1:
@@ -109,7 +110,7 @@ def weighted(loads):
 
 def leastload(loads):
     """Always choose the lowest load.
-    
+
     If the lowest load occurs more than once, the first
     occurance will be used.  If loads has LRU ordering, this means
     the LRU of those with the lowest load is chosen.
@@ -119,32 +120,74 @@ def leastload(loads):
 #---------------------------------------------------------------------
 # Classes
 #---------------------------------------------------------------------
+
+
 # store empty default dependency:
 MET = Dependency([])
 
+
+class Job(object):
+    """Simple container for a job"""
+    def __init__(self, msg_id, raw_msg, idents, msg, header, metadata,
+                    targets, after, follow, timeout):
+        self.msg_id = msg_id
+        self.raw_msg = raw_msg
+        self.idents = idents
+        self.msg = msg
+        self.header = header
+        self.metadata = metadata
+        self.targets = targets
+        self.after = after
+        self.follow = follow
+        self.timeout = timeout
+        self.removed = False # used for lazy-delete from sorted queue
+        
+        self.timestamp = time.time()
+        self.blacklist = set()
+
+    def __lt__(self, other):
+        return self.timestamp < other.timestamp
+
+    def __cmp__(self, other):
+        return cmp(self.timestamp, other.timestamp)
+
+    @property
+    def dependents(self):
+        return self.follow.union(self.after)
+
 class TaskScheduler(SessionFactory):
     """Python TaskScheduler object.
-    
+
     This is the simplest object that supports msg_id based
     DAG dependencies. *Only* task msg_ids are checked, not
     msg_ids of jobs submitted via the MUX queue.
-    
+
     """
-    
-    hwm = Int(0, config=True, shortname='hwm',
+
+    hwm = Integer(1, config=True,
         help="""specify the High Water Mark (HWM) for the downstream
         socket in the Task scheduler. This is the maximum number
-        of allowed outstanding tasks on each engine."""
+        of allowed outstanding tasks on each engine.
+        
+        The default (1) means that only one task can be outstanding on each
+        engine.  Setting TaskScheduler.hwm=0 means there is no limit, and the
+        engines continue to be assigned tasks while they are working,
+        effectively hiding network latency behind computation, but can result
+        in an imbalance of work when submitting many heterogenous tasks all at
+        once.  Any positive value greater than one is a compromise between the
+        two.
+
+        """
     )
     scheme_name = Enum(('leastload', 'pure', 'lru', 'plainrandom', 'weighted', 'twobin'),
-        'leastload', config=True, shortname='scheme', allow_none=False,
+        'leastload', config=True, allow_none=False,
         help="""select the task scheduler scheme  [default: Python LRU]
         Options are: 'pure', 'lru', 'plainrandom', 'weighted', 'twobin','leastload'"""
     )
     def _scheme_name_changed(self, old, new):
         self.log.debug("Using scheme %r"%new)
         self.scheme = globals()[new]
-    
+
     # input arguments:
     scheme = Instance(FunctionType) # function for determining the destination
     def _scheme_default(self):
@@ -153,12 +196,16 @@ class TaskScheduler(SessionFactory):
     engine_stream = Instance(zmqstream.ZMQStream) # engine-facing stream
     notifier_stream = Instance(zmqstream.ZMQStream) # hub-facing sub stream
     mon_stream = Instance(zmqstream.ZMQStream) # hub-facing pub stream
-    
+    query_stream = Instance(zmqstream.ZMQStream) # hub-facing DEALER stream
+
     # internals:
+    queue = Instance(deque) # sorted list of Jobs
+    def _queue_default(self):
+        return deque()
+    queue_map = Dict() # dict by msg_id of Jobs (for O(1) access to the Queue)
     graph = Dict() # dict by msg_id of [ msg_ids that depend on key ]
     retries = Dict() # dict by msg_id of retries remaining (non-neg ints)
     # waiting = List() # list of msg_ids ready to run, but haven't due to HWM
-    depending = Dict() # dict by msg_id of (msg_id, raw_msg, after, follow)
     pending = Dict() # dict by engine_uuid of submitted tasks
     completed = Dict() # dict by engine_uuid of completed tasks
     failed = Dict() # dict by engine_uuid of failed tasks
@@ -171,38 +218,59 @@ class TaskScheduler(SessionFactory):
     all_failed = Set() # set of all failed tasks
     all_done = Set() # set of all finished tasks=union(completed,failed)
     all_ids = Set() # set of all submitted task IDs
-    blacklist = Dict() # dict by msg_id of locations where a job has encountered UnmetDependency
-    auditor = Instance('zmq.eventloop.ioloop.PeriodicCallback')
-    
+
     ident = CBytes() # ZMQ identity. This should just be self.session.session
                      # but ensure Bytes
     def _ident_default(self):
-        return asbytes(self.session.session)
-    
+        return self.session.bsession
+
     def start(self):
+        self.query_stream.on_recv(self.dispatch_query_reply)
+        self.session.send(self.query_stream, "connection_request", {})
+        
         self.engine_stream.on_recv(self.dispatch_result, copy=False)
+        self.client_stream.on_recv(self.dispatch_submission, copy=False)
+
         self._notification_handlers = dict(
             registration_notification = self._register_engine,
             unregistration_notification = self._unregister_engine
         )
         self.notifier_stream.on_recv(self.dispatch_notification)
-        self.auditor = ioloop.PeriodicCallback(self.audit_timeouts, 2e3, self.loop) # 1 Hz
-        self.auditor.start()
-        self.log.info("Scheduler started [%s]"%self.scheme_name)
-    
+        self.log.info("Scheduler started [%s]" % self.scheme_name)
+
     def resume_receiving(self):
         """Resume accepting jobs."""
         self.client_stream.on_recv(self.dispatch_submission, copy=False)
-    
+
     def stop_receiving(self):
         """Stop accepting jobs while there are no engines.
         Leave them in the ZMQ queue."""
         self.client_stream.on_recv(None)
-    
+
     #-----------------------------------------------------------------------
     # [Un]Registration Handling
     #-----------------------------------------------------------------------
     
+    
+    def dispatch_query_reply(self, msg):
+        """handle reply to our initial connection request"""
+        try:
+            idents,msg = self.session.feed_identities(msg)
+        except ValueError:
+            self.log.warn("task::Invalid Message: %r",msg)
+            return
+        try:
+            msg = self.session.unserialize(msg)
+        except ValueError:
+            self.log.warn("task::Unauthorized message from: %r"%idents)
+            return
+        
+        content = msg['content']
+        for uuid in content.get('engines', {}).values():
+            self._register_engine(cast_bytes(uuid))
+
+    
+    @util.log_errors
     def dispatch_notification(self, msg):
         """dispatch register/unregister events."""
         try:
@@ -211,22 +279,22 @@ class TaskScheduler(SessionFactory):
             self.log.warn("task::Invalid Message: %r",msg)
             return
         try:
-            msg = self.session.unpack_message(msg)
+            msg = self.session.unserialize(msg)
         except ValueError:
             self.log.warn("task::Unauthorized message from: %r"%idents)
             return
-        
-        msg_type = msg['msg_type']
-        
+
+        msg_type = msg['header']['msg_type']
+
         handler = self._notification_handlers.get(msg_type, None)
         if handler is None:
             self.log.error("Unhandled message type: %r"%msg_type)
         else:
             try:
-                handler(asbytes(msg['content']['queue']))
+                handler(cast_bytes(msg['content']['uuid']))
             except Exception:
-                self.log.error("task::Invalid notification msg: %r",msg)
-    
+                self.log.error("task::Invalid notification msg: %r", msg, exc_info=True)
+
     def _register_engine(self, uid):
         """New engine with ident `uid` became available."""
         # head of the line:
@@ -237,8 +305,7 @@ class TaskScheduler(SessionFactory):
         self.completed[uid] = set()
         self.failed[uid] = set()
         self.pending[uid] = {}
-        if len(self.targets) == 1:
-            self.resume_receiving()
+
         # rescan the graph:
         self.update_graph(None)
 
@@ -246,11 +313,11 @@ class TaskScheduler(SessionFactory):
         """Existing engine with ident `uid` became unavailable."""
         if len(self.targets) == 1:
             # this was our only engine
-            self.stop_receiving()
-        
+            pass
+
         # handle any potentially finished tasks:
         self.engine_stream.flush()
-        
+
         # don't pop destinations, because they might be used later
         # map(self.destinations.pop, self.completed.pop(uid))
         # map(self.destinations.pop, self.failed.pop(uid))
@@ -259,7 +326,7 @@ class TaskScheduler(SessionFactory):
         idx = self.targets.index(uid)
         self.targets.pop(idx)
         self.loads.pop(idx)
-        
+
         # wait 5 seconds before cleaning up pending jobs, since the results might
         # still be incoming
         if self.pending[uid]:
@@ -269,7 +336,7 @@ class TaskScheduler(SessionFactory):
             self.completed.pop(uid)
             self.failed.pop(uid)
 
-    
+
     def handle_stranded_tasks(self, engine):
         """Deal with jobs resident in an engine that died."""
         lost = self.pending[engine]
@@ -278,7 +345,7 @@ class TaskScheduler(SessionFactory):
                 # prevent double-handling of messages
                 continue
 
-            raw_msg = lost[msg_id][0]
+            raw_msg = lost[msg_id].raw_msg
             idents,msg = self.session.feed_identities(raw_msg, copy=False)
             parent = self.session.unpack(msg[1].bytes)
             idents = [engine, idents[0]]
@@ -288,7 +355,13 @@ class TaskScheduler(SessionFactory):
                 raise error.EngineError("Engine %r died while running task %r"%(engine, msg_id))
             except:
                 content = error.wrap_exception()
-            msg = self.session.msg('apply_reply', content, parent=parent, subheader={'status':'error'})
+            # build fake metadata
+            md = dict(
+                status=u'error',
+                engine=engine,
+                date=datetime.now(),
+            )
+            msg = self.session.msg('apply_reply', content, parent=parent, metadata=md)
             raw_reply = map(zmq.Message, self.session.serialize(msg, ident=idents))
             # and dispatch it
             self.dispatch_result(raw_reply)
@@ -296,41 +369,45 @@ class TaskScheduler(SessionFactory):
         # finally scrub completed/failed lists
         self.completed.pop(engine)
         self.failed.pop(engine)
-    
-    
+
+
     #-----------------------------------------------------------------------
     # Job Submission
     #-----------------------------------------------------------------------
+    
+
+    @util.log_errors
     def dispatch_submission(self, raw_msg):
         """Dispatch job submission to appropriate handlers."""
         # ensure targets up to date:
         self.notifier_stream.flush()
         try:
             idents, msg = self.session.feed_identities(raw_msg, copy=False)
-            msg = self.session.unpack_message(msg, content=False, copy=False)
+            msg = self.session.unserialize(msg, content=False, copy=False)
         except Exception:
             self.log.error("task::Invaid task msg: %r"%raw_msg, exc_info=True)
             return
-        
-        
+
+
         # send to monitor
         self.mon_stream.send_multipart([b'intask']+raw_msg, copy=False)
-        
+
         header = msg['header']
+        md = msg['metadata']
         msg_id = header['msg_id']
         self.all_ids.add(msg_id)
-        
+
         # get targets as a set of bytes objects
         # from a list of unicode objects
-        targets = header.get('targets', [])
-        targets = map(asbytes, targets)
+        targets = md.get('targets', [])
+        targets = map(cast_bytes, targets)
         targets = set(targets)
-            
-        retries = header.get('retries', 0)
+
+        retries = md.get('retries', 0)
         self.retries[msg_id] = retries
-        
+
         # time dependencies
-        after = header.get('after', None)
+        after = md.get('after', None)
         if after:
             after = Dependency(after)
             if after.all:
@@ -352,16 +429,22 @@ class TaskScheduler(SessionFactory):
                 after = MET
         else:
             after = MET
-        
+
         # location dependencies
-        follow = Dependency(header.get('follow', []))
-        
+        follow = Dependency(md.get('follow', []))
+
         # turn timeouts into datetime objects:
-        timeout = header.get('timeout', None)
+        timeout = md.get('timeout', None)
         if timeout:
-            timeout = datetime.now() + timedelta(0,timeout,0)
-        
-        args = [raw_msg, targets, after, follow, timeout]
+            timeout = time.time() + float(timeout)
+
+        job = Job(msg_id=msg_id, raw_msg=raw_msg, idents=idents, msg=msg,
+                 header=header, targets=targets, after=after, follow=follow,
+                 timeout=timeout, metadata=md,
+        )
+        if timeout:
+            # schedule timeout callback
+            self.loop.add_timeout(timeout, lambda : self.job_timeout(job))
         
         # validate and reduce dependencies:
         for dep in after,follow:
@@ -369,66 +452,86 @@ class TaskScheduler(SessionFactory):
                 continue
             # check valid:
             if msg_id in dep or dep.difference(self.all_ids):
-                self.depending[msg_id] = args
+                self.queue_map[msg_id] = job
                 return self.fail_unreachable(msg_id, error.InvalidDependency)
             # check if unreachable:
             if dep.unreachable(self.all_completed, self.all_failed):
-                self.depending[msg_id] = args
+                self.queue_map[msg_id] = job
                 return self.fail_unreachable(msg_id)
-        
+
         if after.check(self.all_completed, self.all_failed):
             # time deps already met, try to run
-            if not self.maybe_run(msg_id, *args):
+            if not self.maybe_run(job):
                 # can't run yet
                 if msg_id not in self.all_failed:
                     # could have failed as unreachable
-                    self.save_unmet(msg_id, *args)
+                    self.save_unmet(job)
         else:
-            self.save_unmet(msg_id, *args)
-    
-    def audit_timeouts(self):
-        """Audit all waiting tasks for expired timeouts."""
-        now = datetime.now()
-        for msg_id in self.depending.keys():
-            # must recheck, in case one failure cascaded to another:
-            if msg_id in self.depending:
-                raw,after,targets,follow,timeout = self.depending[msg_id]
-                if timeout and timeout < now:
-                    self.fail_unreachable(msg_id, error.TaskTimeout)
-                
+            self.save_unmet(job)
+
+    def job_timeout(self, job):
+        """callback for a job's timeout.
+        
+        The job may or may not have been run at this point.
+        """
+        now = time.time()
+        if job.timeout >= (now + 1):
+            self.log.warn("task %s timeout fired prematurely: %s > %s",
+                job.msg_id, job.timeout, now
+            )
+        if job.msg_id in self.queue_map:
+            # still waiting, but ran out of time
+            self.log.info("task %r timed out", job.msg_id)
+            self.fail_unreachable(job.msg_id, error.TaskTimeout)
+
     def fail_unreachable(self, msg_id, why=error.ImpossibleDependency):
         """a task has become unreachable, send a reply with an ImpossibleDependency
         error."""
-        if msg_id not in self.depending:
-            self.log.error("msg %r already failed!", msg_id)
+        if msg_id not in self.queue_map:
+            self.log.error("task %r already failed!", msg_id)
             return
-        raw_msg,targets,after,follow,timeout = self.depending.pop(msg_id)
-        for mid in follow.union(after):
+        job = self.queue_map.pop(msg_id)
+        # lazy-delete from the queue
+        job.removed = True
+        for mid in job.dependents:
             if mid in self.graph:
                 self.graph[mid].remove(msg_id)
-        
-        # FIXME: unpacking a message I've already unpacked, but didn't save:
-        idents,msg = self.session.feed_identities(raw_msg, copy=False)
-        header = self.session.unpack(msg[1].bytes)
-        
+
         try:
             raise why()
         except:
             content = error.wrap_exception()
-        
+        self.log.debug("task %r failing as unreachable with: %s", msg_id, content['ename'])
+
         self.all_done.add(msg_id)
         self.all_failed.add(msg_id)
-        
-        msg = self.session.send(self.client_stream, 'apply_reply', content, 
-                                                parent=header, ident=idents)
-        self.session.send(self.mon_stream, msg, ident=[b'outtask']+idents)
-        
+
+        msg = self.session.send(self.client_stream, 'apply_reply', content,
+                                                parent=job.header, ident=job.idents)
+        self.session.send(self.mon_stream, msg, ident=[b'outtask']+job.idents)
+
         self.update_graph(msg_id, success=False)
-    
-    def maybe_run(self, msg_id, raw_msg, targets, after, follow, timeout):
+
+    def available_engines(self):
+        """return a list of available engine indices based on HWM"""
+        if not self.hwm:
+            return range(len(self.targets))
+        available = []
+        for idx in range(len(self.targets)):
+            if self.loads[idx] < self.hwm:
+                available.append(idx)
+        return available
+
+    def maybe_run(self, job):
         """check location dependencies, and run if they are met."""
-        blacklist = self.blacklist.setdefault(msg_id, set())
-        if follow or targets or blacklist or self.hwm:
+        msg_id = job.msg_id
+        self.log.debug("Attempting to assign task %s", msg_id)
+        available = self.available_engines()
+        if not available:
+            # no engines, definitely can't run
+            return False
+        
+        if job.follow or job.targets or job.blacklist or self.hwm:
             # we need a can_run filter
             def can_run(idx):
                 # check hwm
@@ -436,56 +539,59 @@ class TaskScheduler(SessionFactory):
                     return False
                 target = self.targets[idx]
                 # check blacklist
-                if target in blacklist:
+                if target in job.blacklist:
                     return False
                 # check targets
-                if targets and target not in targets:
+                if job.targets and target not in job.targets:
                     return False
                 # check follow
-                return follow.check(self.completed[target], self.failed[target])
-            
-            indices = filter(can_run, range(len(self.targets)))
+                return job.follow.check(self.completed[target], self.failed[target])
+
+            indices = filter(can_run, available)
 
             if not indices:
                 # couldn't run
-                if follow.all:
+                if job.follow.all:
                     # check follow for impossibility
                     dests = set()
                     relevant = set()
-                    if follow.success:
+                    if job.follow.success:
                         relevant = self.all_completed
-                    if follow.failure:
+                    if job.follow.failure:
                         relevant = relevant.union(self.all_failed)
-                    for m in follow.intersection(relevant):
+                    for m in job.follow.intersection(relevant):
                         dests.add(self.destinations[m])
                     if len(dests) > 1:
-                        self.depending[msg_id] = (raw_msg, targets, after, follow, timeout)
+                        self.queue_map[msg_id] = job
                         self.fail_unreachable(msg_id)
                         return False
-                if targets:
+                if job.targets:
                     # check blacklist+targets for impossibility
-                    targets.difference_update(blacklist)
-                    if not targets or not targets.intersection(self.targets):
-                        self.depending[msg_id] = (raw_msg, targets, after, follow, timeout)
+                    job.targets.difference_update(job.blacklist)
+                    if not job.targets or not job.targets.intersection(self.targets):
+                        self.queue_map[msg_id] = job
                         self.fail_unreachable(msg_id)
                         return False
                 return False
         else:
             indices = None
-            
-        self.submit_task(msg_id, raw_msg, targets, follow, timeout, indices)
+
+        self.submit_task(job, indices)
         return True
-            
-    def save_unmet(self, msg_id, raw_msg, targets, after, follow, timeout):
+
+    def save_unmet(self, job):
         """Save a message for later submission when its dependencies are met."""
-        self.depending[msg_id] = [raw_msg,targets,after,follow,timeout]
+        msg_id = job.msg_id
+        self.log.debug("Adding task %s to the queue", msg_id)
+        self.queue_map[msg_id] = job
+        self.queue.append(job)
         # track the ids in follow or after, but not those already finished
-        for dep_id in after.union(follow).difference(self.all_done):
+        for dep_id in job.after.union(job.follow).difference(self.all_done):
             if dep_id not in self.graph:
                 self.graph[dep_id] = set()
             self.graph[dep_id].add(msg_id)
-    
-    def submit_task(self, msg_id, raw_msg, targets, follow, timeout, indices=None):
+
+    def submit_task(self, job, indices=None):
         """Submit a task to any of a subset of our targets."""
         if indices:
             loads = [self.loads[i] for i in indices]
@@ -498,24 +604,27 @@ class TaskScheduler(SessionFactory):
         # print (target, map(str, msg[:3]))
         # send job to the engine
         self.engine_stream.send(target, flags=zmq.SNDMORE, copy=False)
-        self.engine_stream.send_multipart(raw_msg, copy=False)
+        self.engine_stream.send_multipart(job.raw_msg, copy=False)
         # update load
         self.add_job(idx)
-        self.pending[target][msg_id] = (raw_msg, targets, MET, follow, timeout)
+        self.pending[target][job.msg_id] = job
         # notify Hub
-        content = dict(msg_id=msg_id, engine_id=target.decode('ascii'))
-        self.session.send(self.mon_stream, 'task_destination', content=content, 
+        content = dict(msg_id=job.msg_id, engine_id=target.decode('ascii'))
+        self.session.send(self.mon_stream, 'task_destination', content=content,
                         ident=[b'tracktask',self.ident])
-        
-    
+
+
     #-----------------------------------------------------------------------
     # Result Handling
     #-----------------------------------------------------------------------
+    
+    
+    @util.log_errors
     def dispatch_result(self, raw_msg):
         """dispatch method for result replies"""
         try:
             idents,msg = self.session.feed_identities(raw_msg, copy=False)
-            msg = self.session.unpack_message(msg, content=False, copy=False)
+            msg = self.session.unserialize(msg, content=False, copy=False)
             engine = idents[0]
             try:
                 idx = self.targets.index(engine)
@@ -527,10 +636,10 @@ class TaskScheduler(SessionFactory):
             self.log.error("task::Invaid result: %r", raw_msg, exc_info=True)
             return
 
-        header = msg['header']
+        md = msg['metadata']
         parent = msg['parent_header']
-        if header.get('dependencies_met', True):
-            success = (header['status'] == 'ok')
+        if md.get('dependencies_met', True):
+            success = (md['status'] == 'ok')
             msg_id = parent['msg_id']
             retries = self.retries[msg_id]
             if not success and retries > 0:
@@ -545,19 +654,18 @@ class TaskScheduler(SessionFactory):
                 self.mon_stream.send_multipart([b'outtask']+raw_msg, copy=False)
         else:
             self.handle_unmet_dependency(idents, parent)
-        
+
     def handle_result(self, idents, parent, raw_msg, success=True):
         """handle a real task result, either success or failure"""
         # first, relay result to client
         engine = idents[0]
         client = idents[1]
-        # swap_ids for XREP-XREP mirror
+        # swap_ids for ROUTER-ROUTER mirror
         raw_msg[:2] = [client,engine]
         # print (map(str, raw_msg[:4]))
         self.client_stream.send_multipart(raw_msg, copy=False)
         # now, update our data structures
         msg_id = parent['msg_id']
-        self.blacklist.pop(msg_id, None)
         self.pending[engine].pop(msg_id)
         if success:
             self.completed[engine].add(msg_id)
@@ -567,30 +675,26 @@ class TaskScheduler(SessionFactory):
             self.all_failed.add(msg_id)
         self.all_done.add(msg_id)
         self.destinations[msg_id] = engine
-        
+
         self.update_graph(msg_id, success)
-        
+
     def handle_unmet_dependency(self, idents, parent):
         """handle an unmet dependency"""
         engine = idents[0]
         msg_id = parent['msg_id']
-        
-        if msg_id not in self.blacklist:
-            self.blacklist[msg_id] = set()
-        self.blacklist[msg_id].add(engine)
-        
-        args = self.pending[engine].pop(msg_id)
-        raw,targets,after,follow,timeout = args
-        
-        if self.blacklist[msg_id] == targets:
-            self.depending[msg_id] = args
+
+        job = self.pending[engine].pop(msg_id)
+        job.blacklist.add(engine)
+
+        if job.blacklist == job.targets:
+            self.queue_map[msg_id] = job
             self.fail_unreachable(msg_id)
-        elif not self.maybe_run(msg_id, *args):
+        elif not self.maybe_run(job):
             # resubmit failed
             if msg_id not in self.all_failed:
                 # put it back in our dependency tree
-                self.save_unmet(msg_id, *args)
-        
+                self.save_unmet(job)
+
         if self.hwm:
             try:
                 idx = self.targets.index(engine)
@@ -599,51 +703,78 @@ class TaskScheduler(SessionFactory):
             else:
                 if self.loads[idx] == self.hwm-1:
                     self.update_graph(None)
-        
-        
-    
+
     def update_graph(self, dep_id=None, success=True):
         """dep_id just finished. Update our dependency
-        graph and submit any jobs that just became runable.
-        
-        Called with dep_id=None to update entire graph for hwm, but without finishing
-        a task.
+        graph and submit any jobs that just became runnable.
+
+        Called with dep_id=None to update entire graph for hwm, but without finishing a task.
         """
         # print ("\n\n***********")
         # pprint (dep_id)
         # pprint (self.graph)
-        # pprint (self.depending)
+        # pprint (self.queue_map)
         # pprint (self.all_completed)
         # pprint (self.all_failed)
         # print ("\n\n***********\n\n")
         # update any jobs that depended on the dependency
-        jobs = self.graph.pop(dep_id, [])
+        msg_ids = self.graph.pop(dep_id, [])
 
         # recheck *all* jobs if
         # a) we have HWM and an engine just become no longer full
         # or b) dep_id was given as None
-        if dep_id is None or self.hwm and any( [ load==self.hwm-1 for load in self.loads ]):
-            jobs = self.depending.keys()
         
-        for msg_id in jobs:
-            raw_msg, targets, after, follow, timeout = self.depending[msg_id]
+        if dep_id is None or self.hwm and any( [ load==self.hwm-1 for load in self.loads ]):
+            jobs = self.queue
+            using_queue = True
+        else:
+            using_queue = False
+            jobs = deque(sorted( self.queue_map[msg_id] for msg_id in msg_ids ))
+        
+        to_restore = []
+        while jobs:
+            job = jobs.popleft()
+            if job.removed:
+                continue
+            msg_id = job.msg_id
             
-            if after.unreachable(self.all_completed, self.all_failed)\
-                    or follow.unreachable(self.all_completed, self.all_failed):
+            put_it_back = True
+            
+            if job.after.unreachable(self.all_completed, self.all_failed)\
+                    or job.follow.unreachable(self.all_completed, self.all_failed):
                 self.fail_unreachable(msg_id)
-            
-            elif after.check(self.all_completed, self.all_failed): # time deps met, maybe run
-                if self.maybe_run(msg_id, raw_msg, targets, MET, follow, timeout):
-                    
-                    self.depending.pop(msg_id)
-                    for mid in follow.union(after):
+                put_it_back = False
+
+            elif job.after.check(self.all_completed, self.all_failed): # time deps met, maybe run
+                if self.maybe_run(job):
+                    put_it_back = False
+                    self.queue_map.pop(msg_id)
+                    for mid in job.dependents:
                         if mid in self.graph:
                             self.graph[mid].remove(msg_id)
+                    
+                    # abort the loop if we just filled up all of our engines.
+                    # avoids an O(N) operation in situation of full queue,
+                    # where graph update is triggered as soon as an engine becomes
+                    # non-full, and all tasks after the first are checked,
+                    # even though they can't run.
+                    if not self.available_engines():
+                        break
+            
+            if using_queue and put_it_back:
+                # popped a job from the queue but it neither ran nor failed,
+                # so we need to put it back when we are done
+                # make sure to_restore preserves the same ordering
+                to_restore.append(job)
+        
+        # put back any tasks we popped but didn't run
+        if using_queue:
+            self.queue.extendleft(to_restore)
     
     #----------------------------------------------------------------------
     # methods to be overridden by subclasses
     #----------------------------------------------------------------------
-    
+
     def add_job(self, idx):
         """Called after self.targets[idx] just got the job with header.
         Override with subclasses.  The default ordering is simple LRU.
@@ -651,21 +782,21 @@ class TaskScheduler(SessionFactory):
         self.loads[idx] += 1
         for lis in (self.targets, self.loads):
             lis.append(lis.pop(idx))
-            
-    
+
+
     def finish_job(self, idx):
         """Called after self.targets[idx] just finished a job.
         Override with subclasses."""
         self.loads[idx] -= 1
-    
 
 
-def launch_scheduler(in_addr, out_addr, mon_addr, not_addr, config=None,
+
+def launch_scheduler(in_addr, out_addr, mon_addr, not_addr, reg_addr, config=None,
                         logname='root', log_url=None, loglevel=logging.DEBUG,
                         identity=b'task', in_thread=False):
-    
+
     ZMQStream = zmqstream.ZMQStream
-    
+
     if config:
         # unwrap dict back into Config
         config = Config(config)
@@ -675,23 +806,29 @@ def launch_scheduler(in_addr, out_addr, mon_addr, not_addr, config=None,
         ctx = zmq.Context.instance()
         loop = ioloop.IOLoop.instance()
     else:
-        # in a process, don't use instance() 
+        # in a process, don't use instance()
         # for safety with multiprocessing
         ctx = zmq.Context()
         loop = ioloop.IOLoop()
-    ins = ZMQStream(ctx.socket(zmq.XREP),loop)
-    ins.setsockopt(zmq.IDENTITY, identity)
+    ins = ZMQStream(ctx.socket(zmq.ROUTER),loop)
+    util.set_hwm(ins, 0)
+    ins.setsockopt(zmq.IDENTITY, identity + b'_in')
     ins.bind(in_addr)
-    
-    outs = ZMQStream(ctx.socket(zmq.XREP),loop)
-    outs.setsockopt(zmq.IDENTITY, identity)
+
+    outs = ZMQStream(ctx.socket(zmq.ROUTER),loop)
+    util.set_hwm(outs, 0)
+    outs.setsockopt(zmq.IDENTITY, identity + b'_out')
     outs.bind(out_addr)
     mons = zmqstream.ZMQStream(ctx.socket(zmq.PUB),loop)
+    util.set_hwm(mons, 0)
     mons.connect(mon_addr)
     nots = zmqstream.ZMQStream(ctx.socket(zmq.SUB),loop)
     nots.setsockopt(zmq.SUBSCRIBE, b'')
     nots.connect(not_addr)
     
+    querys = ZMQStream(ctx.socket(zmq.DEALER),loop)
+    querys.connect(reg_addr)
+    
     # setup logging.
     if in_thread:
         log = Application.instance().log
@@ -700,9 +837,10 @@ def launch_scheduler(in_addr, out_addr, mon_addr, not_addr, config=None,
             log = connect_logger(logname, ctx, log_url, root="scheduler", loglevel=loglevel)
         else:
             log = local_logger(logname, loglevel)
-    
+
     scheduler = TaskScheduler(client_stream=ins, engine_stream=outs,
                             mon_stream=mons, notifier_stream=nots,
+                            query_stream=querys,
                             loop=loop, log=log,
                             config=config)
     scheduler.start()
@@ -710,5 +848,5 @@ def launch_scheduler(in_addr, out_addr, mon_addr, not_addr, config=None,
         try:
             loop.start()
         except KeyboardInterrupt:
-            print ("interrupted, exiting...", file=sys.__stderr__)
+            scheduler.log.critical("Interrupted, exiting...")
 
